/* Fetch list of newsgroups */
call RxFuncAdd 'SysLoadFuncs','RexxUtil','SysLoadFuncs'
call SysLoadFuncs
call RxFuncAdd 'SockLoadFuncs','RxSock','SockLoadFuncs'
signal on SYNTAX
nosock=1
call SockLoadFuncs 'q'
nosock=0
signal off SYNTAX

.local['RXIVERSION']=0.15
parse version rxversion
if word(rxversion,1)\='OBJREXX' then do
    say ''
    say 'NGFETCH.CMD requires Object REXX to function.'
    say ''
    say 'You are running: '||rxversion
    say ''
    say 'README.TXT contains more information.'
    exit 4
end

signal on HALT

parse arg hostport switch

if hostport=''|hostport='?'|hostport='/?'|hostport='-?' then do
    say ''
    say 'News group list fetcher v'||.rxiversion
    say ''
    say 'Usage: ngfetch.cmd <host> [/raw] [/q]'
    say ''
    say 'host - hostname of news server'
    say ''
    say '/raw - write all received info without parsing'
    say ''
    say '/q - operate in quiet mode'
    exit
end

if switch~translate~pos('/Q')>0 then quietmode=.true
else quietmode=.false
if switch~translate~pos('/RAW')>0 then .local['WRITERAW']=.true
else .local['WRITERAW']=.false

.local['CRLF']='0d0a'x
.local['BLOCKSIZE']=10240
.local['STINTERVAL']=25
call SysCurState 'off'
.local['CON']=.stream~new('stdout')
listend='0d0a2e0d0a'x
lgroupm='2e0d0a'x
exitcode=0
.local['XFS']='XFER_FINISHED'
.local['STFS']='STATUS_THREAD_EXIT'

outfile=.stream~new('newsrc')

parse var hostport host ':' port
if port='' then port=119

sendstring='LIST'||.crlf

csock=.socket~new

call SysCls
.con~~charout(.crlf~copies(5))~close
.con~~charout('Connecting to host '||host||' on port '||port||'... ')~close
check=csock~connect(host||':'||port)
if check='HALT' then signal HALT
if check=-1 then do
    .con~~charout('Failed!'||.crlf||.crlf)~close
    csock~shutdown('both')
    csock~close
    exit 1
end
else .con~~charout('Done'||.crlf)~close

total=0
data=''
groups=0
check=csock~receive(.blocksize)
if check='HALT' then signal HALT
total=total+check[1]~length
if check[1]~word(1)='400' then do
    say ''
    say 'News server disconnected (probably exceeded per host connection limit).'
    say ''
    csock~shutdown('both')
    csock~close
    exit 2
end
else do
    outfile~open('write replace')
    if .writeraw then outfile~charout(check[1])
end

.con~~charout('Requesting group list... ')~close
check=csock~send(sendstring)
if check='HALT' then signal HALT
if check=-1 then do
    .con~~charout('Failed!'||.crlf||.crlf)~close
    csock~shutdown('both')
    csock~close
    exit 1
end
else .con~~charout('Done'||.crlf)~close

.con~~charout('Receiving group list... '||.crlf||.crlf)~close
call time 'r'

.local['SQ']=.queue~new
.local['TQ']=.queue~new
.local['GQ']=.queue~new
scs=.scrstatus~new
.sq~queue(total)
.gq~queue(groups)

if \quietmode then scs~template
if \quietmode then scs~begin

do until check[1]~right(5)=listend
    check=csock~receive(.blocksize)
    if check='HALT' then signal HALT
    rvar=check[2]
    total=total+check[1]~length
    .sq~queue(total)
    data=data||check[1]
    if writeraw=1 then do
        outfile~charout(data)
        data=''
    end
    else do
        if data~word(1)='215' then do
            parse var data '215' . (.crlf) data
        end
        do while data~pos(.crlf)>0
            parse var data group ' ' (.crlf) data
            outfile~lineout(group)
            groups=groups+1
            .gq~queue(groups)
            if data~left(3)=lgroupm then leave
        end        
    end
end 
.sq~queue(total)
.gq~queue(groups)
.sq~queue(.xfs)
ct=time('e')
stcheck=''
do until stcheck=.stfs
    stcheck=.tq~pull
end
rate=total%ct
outfile~close 
if quietmode\=1 then do
    say ''
    say ''
end
say codel(total)||' bytes transferred in '||codel(ct~format(,2))||' seconds ('||,
    codel(rate)||' bytes/sec.)'
say ''
if writeraw\=1 then do
    say codel(groups)||' groups received.'
    say ''
end

csock~shutdown('both')
csock~close
exit 0

HALT:

if datatype(grow,'NUM')=1 then call SysCurPos grow,70
say ''
say ''
say 'Aborting!'
if csock~objectname='a socket' then do
    csock~shutdown('both')
    csock~close
end
exit 3

SYNTAX:

if nosock=1 then do
    say ''
    say 'Failed to load REXX Socket functions.'
    say ''
    say 'Program cannot continue.'
end
exit 4

/* Socket class - work in progress **********************************/
/*                                                                  */
::class socket

::method init
expose portnum
use arg type,proto
if symbol('type')='LIT' then type='SOCK_STREAM'
if symbol('proto')='LIT' then proto=0
if type~datatype('num')=1 then portnum=type
else portnum=SockSocket('AF_INET',type,proto)
return

::method sock
expose portnum
return portnum

::method connect
use arg servname
signal on HALT
parse var servname hostname ':' host.!port
parse var hostname o1 '.' o2 '.' o3 '.' o4
if o1~datatype='NUM' & o2~datatype='NUM' & o3~datatype='NUM' &,
   o4~datatype='NUM' then do
    if o1~datatype('w')=1 & o2~datatype('w')=1 & o3~datatype('w')=1 &,
       o4~datatype('w')=1 then do
        if (o1>=0 & o1<=255) & (o2>=0 & o2<=255) & (o3>=0 & o3<=255) &,
           (o4>=0 & o4<=255) then do
            stype='IP'
            host.!addr=hostname
        end
    end
end
else stype='NAME'            
if stype='NAME' then call SockGetHostByName hostname,'host.!'
host.!family='AF_INET'
rc=SockConnect(self~sock,'host.!')
return rc
HALT:
return 'HALT'

::method send
use arg sendstring
signal on HALT
rc=SockSend(self~sock,sendstring)
return rc
HALT:
return 'HALT'

::method receive
use arg length
signal on HALT
if length='LENGTH' then length=1024
check=SockRecv(self~sock,'sockdata',length)
return .array~new~~put(sockdata,1)~~put(check,2)
HALT:
return 'HALT'

::method shutdown
use arg how
select
    when how~translate='FROM' then how=0
    when how~translate='TO' then how=1
    otherwise how=2
end
call SockShutDown self~sock,how
return

::method close
call SockClose self~sock
return
/*                                                                  */
/********************************************************************/

/* Screen status object *********************************************/
/*																	*/
::class scrstatus

::method init
expose row srow col
return

::method template
expose row srow col
parse value SysCurPos() with row col
srow=row+2
call SysCurPos row,0
.con~~charout('Bytes stored: ')~close
call SysCurPos row,32
.con~~charout('Bytes/second: ')~~charout(.crlf||.crlf)~close
call SysCurPos srow,0
if \.writeraw then do
    call SysCurPos srow,5
    .con~charout('Groups received: ')
    .con~close
end     
return

::method begin
expose srow
reply
lastt=0.01
currt=0.01
tdiff=0.01
lastb=-1
lastg=0
currb=0
ctotal=0
gtotal=0
stdone=.false
do until stdone
    do while .sq~items>0
        ctotal=.sq~pull
        if ctotal=.xfs then do
            stdone=.true
            ctotal=lntotal
        end
        else lntotal=ctotal           
    end
    do while .gq~items>0
        gtotal=.gq~pull
    end
    if time('e')<0.1 then iterate
    if ctotal>lastb then self~stored(ctotal)
    currt=time('e')
    if currt=0 then currt=0.01
    currb=ctotal-lastb
    tdiff=currt-lastt
    if tdiff=0 then tdiff=0.01
    crate=currb%(tdiff)
    self~rate(crate)
    lastt=currt
    lastb=ctotal
    if gtotal>lastg then self~groups(gtotal)
    lastg=gtotal
    call SysSleep (.stinterval/100)
end
call SysCurPos srow,0
.tq~queue(.stfs)
return
    
::method stored
expose row srow col
use arg bstored
call SysCurPos row,14
.con~~charout(codel(bstored)~left(12))~close
return

::method rate
expose row srow col
use arg rval
call SysCurPos row,46
.con~~charout(codel(rval)~left(12))~close
return

::method groups
expose row srow col
use arg gval
call SysCurPos srow,23
.con~~charout(codel(gval)~left(8))~close
return
/*                                                                  */
/********************************************************************/

/* Comma deliminator routine ****************************************/
/*																	*/
::routine codel
use arg RNum
if RNum~length<3 then return RNum
if RNum~pos('.')>0 then do
    parse var RNum RNum '.' Fract
    Fract='.'||Fract
end
else Fract=''
RNum=RNum~reverse
FNum=''
do while RNum~length>3
	parse var RNum TriDig +3 RNum
	FNum=FNum||TriDig||','
end
FNum=FNum||RNum
return FNum~reverse||Fract
/*																	*/
/********************************************************************/
