/* Fetch documents via HTTP or FTP by URL */
call RxFuncAdd 'SysLoadFuncs','RexxUtil','SysLoadFuncs'
call SysLoadFuncs
call RxFuncAdd 'SockLoadFuncs','RxSock','SockLoadFuncs'
signal on SYNTAX
nosock=1
call SockLoadFuncs 'q'
nosock=0
signal off SYNTAX

numeric digits 10
parse version rxversion
if word(rxversion,1)\='OBJREXX' then do
	say ''
	say 'HTTPGET.CMD requires Object REXX to function.'
	say ''
	say 'You are running: '||rxversion
	say ''
	say 'README.TXT contains more information.'
    exit 6
end
rxversion=rxversion~word(2)

.local['RXIVERSION']=0.15
/* The following is for writing quick debug messages */
/*
.local['DEBUG']=.stream~new('debug.log')
*/
exitval=-1
probtype=0
success=0
failure=0
sockfail=0
servfail=0

signal on HALT
signal on SYNTAX
signal on ERROR
signal on NOVALUE
signal on FAILURE

.local['XFS']='XFER_FINISHED'
.local['STFS']='STATUS_THREAD_EXIT'
parse source progpath
progpath=progpath~subword(3)
pdrive=filespec('drive',progpath)
ppath=filespec('path',progpath)
inifile=.stream~new(pdrive||ppath||'urlget.ini')
getem=.false
getfw=.false
getlog=.false
emadd=''
firewall=''
.local['LOGGING']=.false

emadd=SysIni(inifile~qualify, 'URLGET', 'EMAIL_ADDRESS')
if emadd='ERROR:' then getem=.true
firewall=SysIni(inifile~qualify, 'URLGET', 'FIREWALL')
if firewall='ERROR:' then getfw=.true
.local['LOGGING']=SysIni(inifile~qualify, 'URLGET', 'LOGGING')
if .logging='ERROR:' then getlog=.true

if getem then do
    ok=5
    do until ok=1
        call SysCls
        call SysCurState 'off'
        if ok=0 then do
            call SysCurPos 10,0
            say 'Invalid e-mail address (must contain "@").'
        end
        call SysCurPos 12,0
        say 'Please enter your e-mail address for anonymous FTP transfers:'
        call SysCurState 'on'
        parse pull emadd
        if emadd~pos('@')=0 then ok=0
        else ok=1
    end
    call SysIni inifile~qualify,'URLGET','EMAIL_ADDRESS',emadd
end
if getfw then do
    ok=5
    do until ok=1
        call SysCls
        call SysCurState 'off'
        if ok=0 then do
            call SysCurPos 10,0
            say 'Answer only Y or N.'
        end
        call SysCurPos 12,0
        say 'Are you connected to the Internet through a firewall or proxy server? (Y/N)'
        ans=SysGetKey('noecho')
        if ans~translate\='Y' & ans~translate\='N' then ok=0
        else do
            ok=1
            if ans~translate='Y' then firewall=.true
            else firewall=.false
        end
    end
    call SysIni inifile~qualify,'URLGET','FIREWALL',firewall
end
if getlog then do
    ok=5
    do until ok=1
        call SysCls
        call SysCurState 'off'
        if ok=0 then do
            call SysCurPos 10,0
            say 'Answer only Y or N.'
        end
        call SysCurPos 12,0
        say 'Enable logging? (Y/N)'
        ans=SysGetKey('noecho')
        if ans~translate\='Y' & ans~translate\='N' then ok=0
        else do
            ok=1
            if ans~translate='Y' then .local['LOGGING']=.true
            else .local['LOGGING']=.false
        end
    end
    call SysIni inifile~qualify,'URLGET','LOGGING',.logging
end

.local['STINTERVAL']=25
.local['BLOCKSIZE']=10240
.local['MAXRETRIES']=1000
.local['COMMANDDELAY']=0.1
.local['OVERWRITE']=.true

.local['NOERROR']=0
.local['SOCKERROR']=1
.local['SERVERROR']=2
.local['PATHERROR']=3
.local['AUTHERROR']=4
.local['NOIMPLEMENT']=5
.local['OTHERERROR']=6
.local['ARGERROR']=7
.local['USERABORT']=8
.local['OPTION']=9

urlstring=''
resume=.false
ignorepassive=.false
forceresume=.false
overwrite=.true
quietmode=.false
addtourl=.false
parse arg argstring
if argstring~strip='' then do
    call usage 
    exit 2
end
if argstring~pos('"')>0 then do
    parse var argstring before '"' urlstring '"' after
    urlstring=urlstring~strip
    argstring=before~strip||' '||after~strip
end
do i=1 to argstring~words
    badvar=.false
    currarg=argstring~word(i)~strip
    select
        when addtourl then do
            urlstring=urlstring||currarg
            addtourl=.false
        end
        when (currarg~left(7)~translate='HTTP://' |,
              currarg~left(6)~translate='FTP://') then do
            urlstring=currarg
            if currarg~right(1)=';' then addtourl=.true
        end
        when currarg~left(1)='@' then do
            parse var currarg '@' urllist
            urllist=urllist~strip
            ulist=.stream~new(urllist)
            if ulist~query('exists')='' then do
                say ''
                say 'URL list file not found!'
                exit 2
            end
            else urlstring='fromfile'
        end
        when (currarg~left(1)='/' | currarg~left(1)='-') then do
            select
                when currarg~right(1)~translate='Q' then quietmode=.true
                when currarg~right(1)~translate='R' then resume=.true
                when currarg~right(1)~translate='N' then overwrite=.false
                when currarg~right(1)~translate='P' then ignorepassive=.true
                when currarg~right(1)~translate='L' then .local['LOGGING']=.true
                when currarg~right(1)~translate='W' then firewall=.true
                when currarg~right(1)~translate='F' then do
                    resume=.true
                    forceresume=.true
                end
                when currarg~left(2)~right(1)~translate='D' then do
                    .local['COMMANDDELAY']=currarg~right(currarg~length-2)
                    if .commanddelay~datatype('num')=0 then badvar=.true
                    else if (.commanddelay<0 | .commanddelay>60) then badvar=.true
                    if badvar then do
                        say ''
                        say 'Invalid command delay, '||.commanddelay||'.'
                        exit 2
                    end
                end
                when currarg~left(2)~right(1)~translate='B' then do
                    .local['BLOCKSIZE']=currarg~right(currarg~length-2)
                    if .blocksize~datatype('W')=0 then badvar=.true
                    else if (.blocksize<512 | .blocksize>65535) then badvar=.true
                    if badvar then do
                        say ''
                        say 'Invalid blocksize, '||.blocksize||'.'
                        exit 2
                    end
                end
                when currarg~left(2)~right(1)~translate='T' then do
                    .local['STINTERVAL']=currarg~right(currarg~length-2)
                    if .blocksize~datatype('W')=0 then badvar=.true
                    else if (.stinterval<1 | .stinterval>100) then badvar=.true
                    if badvar then do
                        say ''
                        say 'Invalid status time interval, '||.stinterval||'.'
                        exit 2
                    end
                end
                when currarg~left(2)~right(1)~translate='M' then do
                    .local['MAXRETRIES']=currarg~right(currarg~length-2)
                    if .maxretries~datatype('W')=0 then badvar=.true
                    else if (.maxretries<0 | .maxretries>9999999999) then badvar=.true
                    if badvar then do
                        say ''
                        say 'Invalid maximum retries, '||.maxretries||'.'
                        exit 2
                    end
                end
                otherwise do
                    call usage 
                    exit 2
                end
            end
        end
        when currarg~left(1)=';' then do
            if currarg~length=1 then do
                addtourl=.true
                urlstring=urlstring||currarg
            end
            else urlstring=urlstring||currarg
        end
        otherwise do
            call usage 
            exit 2
        end
    end
end
    
urls=.queue~new

if urlstring='fromfile' then do 
    do while ulist~lines>0
        urls~queue(ulist~linein)
    end
    ulist~close
end
else urls~queue(urlstring)
urls=urls~makearray
	
currdrive=directory()~left(2)
fatcheck=SysFileSystemType(currdrive)

srow=-1
.local['CRLF']='0d0a'x
ncsabug1='NCSA/1.5'||'0a'x
ncsabug2='NCSA/1.5.1'||'0a'x
call SysCurState 'off'
.local['LOG']=.logfile~new(pdrive||ppath||'\urlget.log')
.local['CON']=.stream~new('stdout')
scs=.scrstatus~new

do i=1 to urls~items
    call SysCls
    passivefail=.false
    .local['SIZE']=0
    .local['SQ']=.queue~new
    .local['TQ']=.queue~new
    currxfail=.false
    retries=0
    .local['RESTART']=0
    newname=0
    headend=.crlf||.crlf
    parse value urls[i] with ustring ';' .
    .local['CURRURL']=ustring
    parse value urls[i] with prot '://' host '/' path ';' lfile
    prot=prot~strip~translate
    if prot\='HTTP' & prot\='FTP' then do
        if .logging then .log~failed(.argerror,'invalid URL')
        if urls~items>1 then do
            failure=1
            iterate i
        end
        else do
            call usage
            exit 2
        end
    end
    host=host~strip
    if host='' | host~pos(' ')>0 then do
        if .logging then .log~failed(.argerror,'invalid URL')
        if urls~items>1 then do
            failure=1
            iterate i
        end
		else do
            call usage
            exit 2
		end
    end
    if prot='HTTP' then do
        parse var host host ':' port
        port=port~strip
        if port='' then port=80
    end
    else do
        if host~pos('@')>0 then do
            parse var host userpass '@' host
            if userpass~pos(':')>0 then parse var userpass user ':' pass
            else do
                if .logging then .log~failed(.argerror,'invalid FTP URL')
                if urls~items>1 then do 
                    failure=1
                    iterate i
                end
                else do
                    say ''
                    say 'Invalid FTP URL supplied.'
                    exit 2
                end
            end
            user=user~strip
            pass=pass~strip
            parse value .currurl with before (pass) after
            .local['CURRURL']=before||'*'~copies(pass~length)||after
            parse var host host ':' port
            port=port~strip
            if port='' then port=21
            host=host~strip
        end
        else do
            user='anonymous'
            pass=emadd
            parse var host host ':' port
            port=port~strip
            if port='' then port=21
            host=host~strip
        end
    end
	if path~pos('/')>0 then do
		rpath=path~reverse
		parse var rpath file '/' dirpath
		file=file~reverse~strip
        dirpath='/'||dirpath~reverse~strip
	end
	else do
		file=path~strip
		dirpath='/'
	end
	lfile=lfile~strip
    if file='' then do
        if lfile='' then do
            if .logging then .log~failed(.argerror,'invalid URL')
            if urls~items>1 then do
                failure=1
                iterate i
            end
            else do
                call usage
                exit 2
            end
        end
    end
	if lfile='' then lfile=file
    if dirpath~pos(' ')>0 & prot='HTTP' then do until dirpath~pos(' ')=0
        parse var dirpath before ' ' after
        dirpath=before||'%20'||after
    end
	longea=0
    dfile=file
    if file~pos(' ')>0 & prot='HTTP' then do until file~pos(' ')=0
        parse var file before ' ' after
        file=before||'%20'||after
    end
    ldrive=filespec('drive',lfile)
    if ldrive\='' then do
        fatcheck=SysFileSystemType(ldrive)
    end
    if fatcheck='FAT' then do
        ldrive=filespec('drive',lfile)
        if ldrive='' then ldrive='DEF'
        lpath=filespec('path',lfile)
        if lpath='' then lpath='DEF'
        lname=filespec('name',lfile)
        parse var lname basename '.' ext
        parse value lname~reverse with ext2 '.' basename2
		select 
			when ext\=ext2~reverse then newname=1
			when basename\=basename2~reverse then newname=1
			when basename~length>8 then newname=1
			when ext~length>3 then newname=1
			when basename~pos(' ')>0 then newname=1
			when ext~pos(' ')>0 then newname=1
            otherwise newname=0
		end
		if newname=1 then do
			if ext~length<=3 then shext='.'||ext
            else shext='.UGF'
		end
		done=0
		if newname=1 then do until done=1
			p1=right(d2x(random(65535)),4,'0')
			p2=right(d2x(random(65535)),4,'0')
            select
                when ldrive='DEF' & lpath='DEF' then ltot=p1||p2||shext
                when ldrive='DEF' then ltot=lpath||p1||p2||shext
                when lpath='DEF' then ltot=ldrive||p1||p2||shext
                otherwise ltot=ldrive||lpath||p1||p2||shext
            end
            outfile=.stream~new(ltot)
			if outfile~query('exists')='' then do
				done=1
				longea=1
                longval='FDFF'x||filespec('name',lfile)~length~d2c||,
                        '00'x||filespec('name',lfile)
                lname=filespec('name',outfile~qualify)
                select
                    when ldrive='DEF' & lpath='DEF' then lfile=lname
                    when ldrive='DEF' then lfile=lpath||lname
                    when lpath='DEF' then lfile=ldrive||lname
                    otherwise lfile=ldrive||lpath||lname
                end
            end
		end
		else outfile=.stream~new(lfile)
	end
	else outfile=.stream~new(lfile)

    if \overwrite & \(prot='FTP' & resume) then do
        if outfile~query('exists')\='' then do
            say ''
            say 'Local file exists, not overwriting.'
            if .logging then .log~failed(.option,'overwrite disabled, file exists')
            iterate i
        end
    end

	servport=host||':'||port

	csock=.socket~new

	say ''
	say ''
    .con~~charout('Connecting to host '||host||' on port '||port||'... ')~close
	check=csock~connect(servport)
	if check='HALT' then signal HALT
	if check=-1 then do
        if .logging then .log~failed(.sockerror,'host connection failed')
        .con~~charout('Failed!'||.crlf)~close
        call sockclean('control')
        sockfail=1
        failure=1
        iterate i
	end
    else .con~~charout('Done'||.crlf)~close

    if prot='HTTP' then httpsendstring='GET '||dirpath||'/'||file||' HTTP/1.0'||.crlf||,
                       'Accept: */*'||.crlf||,
                       'Accept-Encoding: */*'||.crlf||,
                       'User-Agent: URLGet/'||.rxiversion||'  Object REXX/'||rxversion||.crlf||,
                       .crlf

    if prot='HTTP' then do
        .con~~charout('Sending request string... ')~close
        check=csock~send(httpsendstring)
        if check='HALT' then signal HALT
        if check=-1 then do
            if .logging then .log~failed(.sockerror,'HTTP send request failed')
            .con~~charout('Failed!'||.crlf)~close
            call sockclean('control')
            sockfail=1
            failure=1
            iterate i            
        end
        else .con~~charout('Done'||.crlf)~close
    end

    else do 
        fs=.FTPSession~new(csock,user,pass,dirpath,file)
        .con~~charout('Logging in... ')~close
        rval=fs~login
        if rval='HALT' then signal HALT
        if rval\=.noerror then do
            failure=1
            .con~~charout('Failed!'||.crlf)~close
            say ''
            select
                when rval=.sockerror then do
                    sockfail=1
                    say 'Socket error logging in.'
                end
                when rval=.serverror then do
                    servfail=1
                    say 'Login not allowed.'
                end
                otherwise nop
            end
            iterate i
        end
        else .con~~charout('Done'||.crlf)~close
        .con~~charout('Setting binary transfer mode... ')~close
        rval=fs~setbinary
        if rval='HALT' then signal HALT
        if rval\=.noerror then do
            .con~~charout('Failed!'||.crlf)~close
            say ''
            say 'Failed to set binary transfer mode.'
            failure=1
            if rval=.sockerror then sockfail=1
            else servfail=1
            iterate i
        end
        else .con~~charout('Done'||.crlf)~close
        .con~~charout('Changing to directory '||dirpath||'... ')~close
        rval=fs~chdir
        if rval='HALT' then signal HALT
        if rval\=.noerror then do
            say ''
            say 'Failed to change working directory.'
            failure=1
            select
                when rval=.sockerror then sockfail=1
                when rval=.patherror then do
                    servfail=1
                    say 'Directory does not exist or access denied.'
                    iterate i
                end
            end
        end
        else .con~~charout('Done'||.crlf)~close
        if file\='' then do
            .con~~charout('Retrieving file size... ')~close
            rval=fs~getsize
            if rval='HALT' then signal HALT
            if rval\=.noerror then do
                select
                    when rval=.sockerror then do
                        sockfail=1
                        failure=1
                        say ''
                        say 'Socket error while getting file size.'
                        iterate i
                    end
                    when rval=.falseimplement then do
                        if forceresume then do
                            servfail=1
                            failure=1
                            if .logging then .log~failed(.serverror,'resume required, cannot determine remote file size')
                            say ''
                            say 'Cannot determine remote size.'
                            iterate i
                        end
                        resume=.false
                        fs~setresume(.false)
                        .local['SIZE']=0
                    end
                    otherwise .con~~charout('Failed!'||.crlf)~close
                end
            end
            else do 
                .local['SIZE']=fs~size 
                .con~~charout('Done'||.crlf)~close
            end
        end
        else .local['size']=0
        if firewall then do
            .con~~charout('Setting passive mode... ')~close
            rval=fs~setpassive
            if rval='HALT' then signal HALT
            if rval\=.noerror then do
                select
                    when rval=.sockerror then do
                        sockfail=1
                        say 'Socket error setting passive mode.'
                    end
                    when rval=.autherror then do
                        if ignorepassive then do
                            passivefail=.true
                            .con~~charout('Failed!'||.crlf)~close
                        end
                        else do
                            failure=1
                            servfail=1
                            if .logging then .log~failed(.serverror,'passive mode required, but denied: '||.passdenycode)
                            say ''
                            say 'Server refused passive connection.'
                            iterate i
                        end
                    end
                    otherwise .con~~charout('Failed!'||.crlf)~close
                end
            end
            else .con~~charout('Done'||.crlf)~close
        end
        if (\firewall | (firewall & passivefail)) then do
            .con~~charout('Sending data port... ')~close
            rval=fs~sendport
            if rval='HALT' then signal HALT
            if rval\=.noerror then do
                failure=1
                say ''
                select
                    when rval=.sockerror then do
                        sockfail=1
                        say 'Socket error sending data port.'
                    end
                    otherwise do
                        servfail=1
                        say 'Unable to send data port.'
                    end
                end
                iterate i
            end
            else .con~~charout('Done'||.crlf)~close
        end
        if resume & file\='' then do
            if outfile~query('exists')\='' then do
                if (.size>0 & .size>outfile~query('size')) then do
                    .con~~charout('Setting restart marker... ')~close
                    rval=fs~setrestart(outfile)
                    if rval='HALT' then signal HALT
                    if rval\=.noerror then do
                        select
                            when rval=.sockerror then do
                                sockfail=1
                                failure=1
                                .con~~charout('Failed!'||.crlf)~close
                                say ''
                                say 'Socket error setting restart marker.'
                                iterate i
                            end
                            otherwise do
                                if forceresume then do
                                    failure=1
                                    servfail=1    
                                    if .logging then .log~failed(.serverror,'resume required, not supported by server')
                                    .con~~charout('Failed!'||.crlf)~close
                                    say ''
                                    say 'Server does not support resume, aborting transfer.'
                                    iterate i
                                end
                                else .con~~charout('Failed!'||.crlf)~close
                            end
                        end
                    end
                    else do
                        .local['RESTART']=fs~restart
                        .con~~charout('Done'||.crlf)~close
                    end
                end
                else do
                    say ''
                    say 'Local file exists, and is not smaller than server copy.'
                    say 'Not overwriting.'
                    if .logging then .log~failed(.option,'overwrite disabled, file exists')
                    iterate i
                end
            end
        end
        .con~~charout('Sending retrieve request... ')~close
        if file='' then rval=fs~retrievelisting
        else rval=fs~retrieve
        if rval='HALT' then signal HALT
        if rval~objectname\='an Array' then do
            failure=1
            say ''
            select
                when rval=.sockerror then do
                    sockfail=1
                    say 'Socket error on retrieve.'
                end
                otherwise do
                    servfail=1
                    say 'Unable to retrieve.'
                end
            end
            iterate i
        end
        else do
            dsock=rval[2]
            if dsock~objectname~translate\='A SOCKET' then do
                failure=1
                if .logging then .log~failed(.othererror,'error with data socket')
                say ''
                say 'Error with data socket.'
                iterate i
            end
            else .con~~charout('Done'||.crlf)~close
        end 
    end        

    if (prot='HTTP' | (prot='FTP' & .restart=0)) then do
        if outfile~open('write replace')\='READY:' then do
            say ''
            say ''
            say 'Unable to open file '||outfile~qualify||' for writing!'
            failure=1
            iterate i
        end
    end
    else do
        if outfile~open('write append')\='READY:' then do
            say ''
            say ''
            say 'Unable to open file '||outfile~qualify||' for writing!'
            failure=1
            iterate i
        end
    end

    total=0
    .sq~queue(total)
    nofile=0     
    headgone=0
    security=0
	dumbserver=0
    .local['FINISHED']=.false

    say 'Receiving URL:'
    say ''
    say .currurl
    say ''
    if lfile\=file then do
        say 'Stored locally as '||lfile||'...'
        say ''
    end
    if (prot='FTP' & .restart>0) then do
        say 'Resuming from byte '||codel(.restart)||...
        say ''
    end

    call time 'r'
    do until .finished
        if prot='HTTP' then check=csock~receive(.blocksize)
        else check=dsock~receive(.blocksize)
		if check='HALT' then signal HALT
        rvar=check[2]
        if rvar<0 then do
            if retries>.maxretries then .local['FINISHED']=.true
			else retries=retries+1
		end
        else if rvar=0 then .local['FINISHED']=.true
        data=check[1]
        if \headgone then do
            if \quietmode then scs~template
            if data~pos(ncsabug1)>0 | data~pos(ncsabug2)>0 then headend='0a0a'x
            if prot='HTTP' & data~word(2)='404' then nofile=1
            else if prot='HTTP' & data~word(2)='403' then security=1
            else if prot='HTTP' & data~word(2)='406' then dumbserver=1
            else if data~pos(headend)>0 | prot='FTP' then do
				headgone=1
                if prot='HTTP' then do
                    parse var data header (headend) data
                    if header~pos('ength:')>0 then do
                        header=header||.crlf
                        parse var header 'ength:' stemp (.crlf)
                        .local['SIZE']=stemp~strip
                        if .size~datatype='NUM' then nop
                        else .local['SIZE']=0
                    end
                    else .local['SIZE']=0
                end
                if .size=0 then do
                    if \quietmode then scs~sizeunknown
                    else do
                        say 'Total size: UNKNOWN'
                        say ''
                    end
                end
                else do
                    if \quietmode then scs~size(.size)
					else do
                        say 'Total size: '||codel(.size)||' bytes'
                        say ''
                    end
                end
				outfile~charout(data)
                total=total+data~length
                .sq~queue(total)
                if \quietmode then scs~begin
			end
		end
		else do
			outfile~charout(data)
            total=total+data~length
            .sq~queue(total)
		end
        if .finished then do
			ct=time('e')
            .sq~queue(total)
            .sq~queue(.xfs)
            stcheck=''
            do until stcheck=.stfs
                stcheck=.tq~pull
            end
            if .size>0 then do
                if (total+.restart)<.size then do
                    if .logging then .log~failed(.sockerror,'maximum retries exceeded')
					say ''
					say ''
					say 'Maximum retries reached - transfer failed!'
                    sockfail=1
                    failure=1
                    currxfail=.true
				end
			end
            rate=total%ct
            if \quietmode then do
				say ''
				say ''
			end
            say codel(total)||' bytes received in '||codel(ct~format(,2))||,
					' seconds ('||codel(rate)||' bytes/sec.)'
            if .logging & \currxfail then do
                if .restart>0 then .log~transfer(total,ct,.restart)
                else .log~transfer(total,ct)
            end
		end
		if nofile=1 then do
            if .logging then .log~failed(.serverror,"remote file doesn't exist")
            say ''
            say ''
            say "Remote file doesn't exist!"
			longea=0
		end
		if dumbserver=1 then do
            if .logging then .log~failed(.serverror,'improper server returned 406')
			say ''
			say ''
			say "Incorrect protocol usage at remote server, returned 406!"
			longea=0
		end
        if security=1 then do
            if .logging then .log~failed(.serverror,'no access or other error')
            say ''
            say ''
            say "No permission or other access error!"
            longea=0
        end
        if nofile=1 | dumbserver=1 | security=1 then leave
	end
    if nofile\=1 then do
        success=1
        outfile~close 
    end
	if longea=1 then call SysPutEA outfile~qualify,'.LONGNAME',longval

    if prot='FTP' then do
        fs~logoff
        dsock~shutdown('both')
        dsock~close
    end
    csock~shutdown('both')
	csock~close
end

select 
    when success=1 & failure=0 then exitval=0
    when success=1 & failure=1 then exitval=1
    when urls~items=1 & sockfail=1 then exitval=3
    when urls~items=1 & servfail=1 then exitval=4
    when urls~items>1 & sockfail=1 & servfail=1 & success=0 then exitval=30
    otherwise exitval=255
end

exit exitval

SYNTAX:
if nosock=1 then do
    say ''
    say 'Failed to lock REXX Socket functions.'
    say ''
    say 'Program cannot continue.'
    exit 6
end
probtype='SYNTAX'
probline=sigl
signal handler
ERROR:
probtype='ERROR'
probline=sigl
signal handler
NOVALUE:
probtype='NOVALUE'
probline=sigl
signal handler
FAILURE:
probtype='FAILURE'
probline=sigl
signal handler
handler:
call SysCls
say ''
say ''
say 'Problem of type '||probtype||' on source line #'||probline||'!'
say ''
if symbol('rc')='VAR' then say 'Cause of error: '||rc||' - '||errortext(rc)
say ''
say 'Source line content:'
say sourceline(probline)
say ''
say 'Please notify thanny@home.com about this, with as much possible detail.'
exitval=6
HALT: 
if symbol('quietmode')='VAR' then if \quietmode then do
    parse value SysCurPos() with row col
    srow=row+4
	call SysCurPos srow,75
end
else do
	say ''
	say ''
end
say ''
say ''
say 'Aborting!'
if symbol('.log')='VAR' & probtype=0 then .log~failed(.userabort,'program halted')
call sockclean
if exitval=-1 then do
    select
        when success=0 then exitval=5
        when success=1 then do
            select
                when failure=0 then exitval=0
                when symbol('urls')='VAR' then do
                    if symbol('i')='VAR' then do
                        if urls~items>=i then exitval=1
                    end
                    else exitval=0
                end
            end
        end
    end
end
.sq~queue(.xfs)
exit exitval

sockclean: procedure expose csock dsock

if symbol('csock')='VAR' then if csock~objectname~translate='A SOCKET' then do
    csock~shutdown('both')
    csock~close
end
if symbol('dsock')='VAR' then if dsock~objectname~translate='A SOCKET' then do
    dsock~shutdown('both')
    dsock~close
end
return

/* Socket class - work in progress **********************************/
/*																	*/
::class socket

::method init
expose portnum
use arg type,proto
if symbol('type')='LIT' then type='SOCK_STREAM'
if symbol('proto')='LIT' then proto=0
if type~datatype('num')=1 then portnum=type
else portnum=SockSocket('AF_INET',type,proto)
return

::method sock
expose portnum
return portnum

::method connect
use arg servname
signal on HALT
parse var servname hostname ':' host.!port
parse var hostname o1 '.' o2 '.' o3 '.' o4
if o1~datatype='NUM' & o2~datatype='NUM' & o3~datatype='NUM' &,
   o4~datatype='NUM' then do
	if o1~datatype('w')=1 & o2~datatype('w')=1 & o3~datatype('w')=1 &,
	   o4~datatype('w')=1 then do
		if (o1>=0 & o1<=255) & (o2>=0 & o2<=255) & (o3>=0 & o3<=255) &,
		   (o4>=0 & o4<=255) then do
			stype='IP'
			host.!addr=hostname
		end
	end
end
else stype='NAME'            
if stype='NAME' then call SockGetHostByName hostname,'host.!'
host.!family='AF_INET'
rc=SockConnect(self~sock,'host.!')
return rc
HALT:
return 'HALT'

::method send
use arg sendstring
signal on HALT
rc=SockSend(self~sock,sendstring)
return rc
HALT:
return 'HALT'

::method receive
use arg length
signal on HALT
if length='LENGTH' then length=1024
check=SockRecv(self~sock,'sockdata',length)
return .array~new~~put(sockdata,1)~~put(check,2)
HALT:
return 'HALT'

::method bind
host.!addr=SockGetHostId()
host.!port=self~sock
host.!family='AF_INET'
rval=SockBind(self~sock,'host.!')
return rval

::method listen
rval=SockListen(self~sock, 65536)
return rval

::method accept
use arg goodaddr
newsock=SockAccept(self~sock)
return newsock

::method shutdown
use arg how
select
	when how~translate='FROM' then how=0
	when how~translate='TO' then how=1
	otherwise how=2
end
call SockShutDown self~sock,how
return

::method close
call SockClose self~sock
return
/*																	*/
/********************************************************************/

/* FTP session object ***********************************************/
/*                                                                  */
::class FTPSession

::method init
expose fsize dirpath user pass file marker dsock lsock csock passivemode
signal on HALT
use arg csock,user,pass,dirpath,file
marker=0
passivemode=.false
return

::method sethost
expose host
signal on HALT
use arg host
return

::method setport
expose port
signal on HALT
use arg port
return

::method setuser
expose user
signal on HALT
use arg user
return

::method setpass
expose pass
signal on HALT
use arg pass
return

::method setfile
expose file
signal on HALT
use arg file
return

::method setpath
expose dirpath
signal on HALT
use arg dirpath
return

::method setresume
expose resume
signal on HALT
use arg resume
return

::method size
expose fsize
signal on HALT
return fsize

::method login
expose user pass
signal on HALT
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'waiting for FTP server')
        return .sockerror
    end
    when rval[2]\='220' then do
        if .logging then .log~failed(.serverror,'not accepting users: '||rval[2])
        return .serverror
    end
    otherwise nop
end
rval=self~send('USER '||user)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending user name')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'    
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'awaiting password request')
        return .sockerror
    end
    when rval[2]\='331' then do
        if .logging then .log~failed(.serverror,'user not allowed: '||rval[2])
        return .serverror
    end
    otherwise nop
end
rval=self~send('PASS '||pass)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending password')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'awaiting password acceptance')
        return .sockerror
    end
    when rval[2]\='230' then do 
        if .logging then .log~failed(.serverror,'password denied: '||rval[2])
        return .autherror
    end
    otherwise return .noerror
end

::method logoff
signal on HALT
rval=self~send('QUIT')
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then return .sockerror
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'    
    when rval[1]=-1 then return .sockerror
    when rval[2]\='226' then .serverror
    otherwise return .noerror
end

::method setbinary
expose file
signal on HALT
rval=self~send('TYPE I')
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending binary mode request')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'awaiting binary mode confirmation')
        return .sockerror
    end
    when rval[2]\='200' then do
        if .logging then .log~failed(.serverror,'could not set binary transfer mode: '||rval[2])
        return .serverror
    end
    otherwise return .noerror
end

::method chdir
expose dirpath file
signal on HALT
rval=self~send('CWD '||dirpath)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending directory change')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'awaiting directory change confirmation')
        return .sockerror
    end
    when (rval[2]='550' | rval[2]='450') then do
        if .logging then .log~failed(.serverror,'bad path or no access: '||rval[2])
        return .patherror
    end
    when rval[2]='250' then return .noerror
    otherwise return .othererror
end

::method setpassive
expose passivemode dsock
signal on HALT
rval=self~send('PASV')
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending passive mode request')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'awaiting passive mode confirmation')
        return .sockerror
    end
    when rval[2]='425' then do
        .local['PASSDENYCODE']=rval[2]
        return .autherror
    end
    when rval[2]='227' then do
        recd=rval[3]
        parse var recd . '(' hostport ')' .
        parse var hostport h1 ',' h2 ',' h3 ',' h4 ',' p1 ',' p2
        host=h1||'.'||h2||'.'||h3||'.'||h4
        port=(p1~d2c||p2~d2c)~c2d
        servport=host||':'||port
        dsock=.socket~new
        check=dsock~connect(servport)                                                                                             
        if check='HALT' then return HALT                                    
        if check=-1 then do
            if .logging then .log~failed(.sockerror,'could not establish data connection')
            return .sockerror
        end
        passivemode=.true
        return .noerror
    end
    otherwise return .othererror
end

::method setrestart
expose resume marker 
signal on HALT
use arg ofile
if ofile='' then return .othererror
if fsize='' | fsize=0 then return .othererror
if ofile~query('exists')\='' then do
    marker=ofile~query('size')
    if fsize<=marker then return .falseimplement
end
rval=self~send('REST '||marker)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending restart marker')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving restart confirmation')
        return .sockerror
    end
    when rval[2]='350' then return .noerror
    otherwise return .falseimplement
end

::method restart
expose marker
signal on HALT
return marker

::method getsize
expose fsize file
signal on HALT
rval=self~send('SIZE '||file)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'requesting file size')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving file size info')
        return .sockerror
    end
    when rval[2]='213' then do
        parse value rval[3] with '213 ' fsize (.crlf)
        fsize=fsize~strip
        if fsize~datatype('W')=0 then fsize=0
    end
    otherwise nop
end
if fsize\=0 then return .noerror        
rval=self~send('STAT '||file)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'requesting file status info')
        return .sockerror
    end        
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving file status info')
        return .sockerror
    end
    when rval[2]='213' then do
        recd=rval[3]
        do until recd~pos(.crlf)=0
            parse var recd begin (.crlf) recd
            if begin~translate~pos(file~translate)>0 then do
                if begin~left(4)=sval||'-' then nop
                else do
                    fsize=begin~word(5)
                    if fsize~datatype('W')=0 then fsize=begin~word(4)
                    recd=''
                end
            end
        end
        if fsize~datatype('W')=0 then fsize=0
    end
    otherwise nop
end
if fsize\=0 then return .noerror
else return .falseimplement
        
::method sendport
expose lsock
signal on HALT
lsock=.socket~new
lsock~bind
lsock~listen
haddr=SockGetHostId()
parse var haddr p1 '.' p2 '.' p3 '.' p4
haddr=p1||','||p2||','||p3||','||p4
lport=lsock~sock
lport=lport~d2x~right(4,'0')
lport=lport~left(2)~x2d||','||lport~right(2)~x2d
pstring=haddr||','||lport
rval=self~send('PORT '||pstring)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending data port')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving data port confirmation')
        return .sockerror
    end
    when rval[2]='200' then return .noerror
    otherwise do
        if .logging then .log~failed(.serverror,'data port not accepted: '||rval[2])
        return .serverror
    end
end

::method retrieve
expose file dsock lsock passivemode
signal on HALT
rval=self~send('RETR '||file)
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending retrieve request')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving retrieve request confirmation')
        return .sockerror
    end
    when (rval[2]='125' | rval[2]='150') then do
        if \passivemode then dsock=.socket~new(lsock~accept)
        ret=.queue~new
        ret~queue(.noerror)
        ret~queue(dsock)
        ret=ret~makearray
        return ret
    end
    otherwise do
        if .logging then .log~failed(.serverror,'retrieve request denied: '||rval[2])
        return .serverror
    end
end

::method retrievelisting
expose file dsock lsock passivemode
signal on HALT
rval=self~send('LIST *')
select
    when rval='HALT' then return 'HALT'
    when rval=-1 then do
        if .logging then .log~failed(.sockerror,'sending retrieve request')
        return .sockerror
    end
    otherwise nop
end
rval=self~receive(.blocksize)
select
    when rval='HALT' then return 'HALT'
    when rval[1]=-1 then do
        if .logging then .log~failed(.sockerror,'receiving retrieve request confirmation')
        return .sockerror
    end
    when (rval[2]='125' | rval[2]='150') then do
        if \passivemode then dsock=.socket~new(lsock~accept)
        ret=.queue~new
        ret~queue(.noerror)
        ret~queue(dsock)
        ret=ret~makearray
        return ret
    end
    otherwise do
        if .logging then .log~failed(.serverror,'retrieve request denied: '||rval[2])
        return .serverror
    end
end

::method send
expose csock
signal on HALT
use arg sendstring
if sendstring='' then return
call SysSleep .commanddelay
check=csock~send(sendstring||.crlf)
return check

::method receive
expose csock
signal on HALT
call SysSleep .commanddelay
check=csock~receive(.blocksize)
ret=.queue~new
if check='HALT' then return 'HALT'
if check[2]=-1 then do
    ret~queue(check[2])
    ret=ret~makearray
    return ret
end
sval=check[1]~left(3)
ret~queue(check[2])
ret~queue(sval)
recd=check[1]
if check[1]~left(4)=sval||'-' then do until done=1
    bend=sval||' '
    parse var recd (bend) stuff
    if stuff\='' & stuff~right(2)=.crlf then leave        
    scrap=check[1]
    call SysSleep .commanddelay
    check=csock~receive(.blocksize)
    if check='HALT' then return HALT
    if check[2]=-1 then return .sockerror
    recd=recd||check[1]
    scrap=scrap||check[1]
    do until scrap~pos(.crlf)=0
        parse var scrap begin (.crlf) scrap
        if begin~left(4)=sval||' ' then pdone=1
        else pdone=0
    end
    if scrap='' & pdone=1 then done=1
end
ret~queue(recd)
ret=ret~makearray
return ret

HALT:
return 'HALT'
/*                                                                  */
/********************************************************************/

/* Log object *******************************************************/
/*                                                                  */
::class logfile

::method init
expose lf filename
use arg filename
if symbol('filename')\='VAR' then filename='urlget.log'
lf=.stream~new(filename)
return

::method transfer
use arg size,time,restart
wstring='['||date()||' - '||time()||'] - Successful transfer for:'||.crlf
wstring=wstring||.currurl||.crlf
wstring=wstring||'Received '||codel(size)||' bytes in '||codel(time~format(,2))||' seconds'
wstring=wstring||' ('||codel(size%time)||' bytes/second).'||.crlf
if symbol('restart')='VAR' then do
    wstring=wstring||'Total size is '||codel(size+restart)||' bytes, resumed '
    wstring=wstring||'from byte '||codel(restart)||'.'||.crlf
end
wstring=wstring||'-'~copies(75)||.crlf
self~write(wstring)
return

::method failed
use arg reason,detail
wstring='['||date()||' - '||time()||'] - Transfer failed for:'||.crlf
wstring=wstring||.currurl||.crlf
wstring=wstring||'Failure cause: '
select
    when reason=.sockerror then wstring=wstring||'socket error'
    when reason=.serverror then wstring=wstring||'server error'
    when reason=.argerror then wstring=wstring||'argument error'
    when reason=.userabort then wstring=wstring||'transfer aborted'
    when reason=.option then wstring=wstring||'option prevents transfer'
    otherwise wstring=wstring||'unknown error'
end
wstring=wstring||' ('||detail||').'||.crlf
wstring=wstring||'-'~copies(75)||.crlf
self~write(wstring)
return

::method write
expose lf
use arg wstring
reply
written=.false
do until written
    if lf~open('write')\='READY:' then iterate
    else do
        lf~charout(wstring)
        lf~close
        written=.true
    end
end
return
/*                                                                  */
/********************************************************************/

/* Screen status object *********************************************/
/*																	*/
::class scrstatus

::method init
expose row srow col
return

::method template
expose row srow col
parse value SysCurPos() with row col
srow=row+2
call SysCurPos row,0
.con~~charout('Bytes stored: ')~close
call SysCurPos row,32
.con~~charout('Bytes/second: ')~~charout(.crlf||.crlf)~close
call SysCurPos srow,0
.con~~charout('Total size: ')~close
call SysCurPos srow,36
.con~~charout('Progress: ')~close
return

::method sizeunknown
expose row srow col
call SysCurPos srow,12
.con~~charout('UNKNOWN')~close
call SysCurPos srow,46
.con~~charout('UNKNOWN')~close
return

::method size
expose row srow col
use arg size
call SysCurPos srow,12
.con~~charout(codel(size)||' bytes')~close                        
call SysCurPos srow,49
.con~~charout('%')~close
return

::method begin
expose srow
reply
lastt=0.01
currt=0.01
tdiff=0.01
lastb=-1
currb=0
ctotal=0
pval=0
lastp=-1
stdone=.false
do until stdone
    do while .sq~items>0
        ctotal=.sq~pull
        if ctotal=.xfs then do
            stdone=.true
            ctotal=lntotal
        end
        else lntotal=ctotal           
    end
    if time('e')<0.1 then iterate
    if ctotal>lastb then self~stored(ctotal+.restart)
    currt=time('e')
    if currt=0 then currt=0.01
    currb=ctotal-lastb
    tdiff=currt-lastt
    if tdiff=0 then tdiff=0.01
    crate=currb%(tdiff)
    self~rate(crate)
    lastt=currt
    lastb=ctotal
    if .size>0 then do
        pval=((ctotal+.restart)/.size*100)~format(,0)
        if pval>lastp then self~percent(pval)
        lastp=pval
    end
    call SysSleep (.stinterval/100)
end
call SysCurPos srow,0
.tq~queue(.stfs)
return
    
::method stored
expose row srow col
use arg bstored
call SysCurPos row,14
.con~~charout(codel(bstored)~left(12))~close
return

::method rate
expose row srow col
use arg rval
call SysCurPos row,46
.con~~charout(codel(rval)~left(12))~close
return

::method percent
expose row srow col
use arg pval
call SysCurPos srow,46
.con~~charout(pval~right(3))~close
return
/*                                                                  */
/********************************************************************/

/* Comma deliminator routine ****************************************/
/*																	*/
::routine codel
use arg RNum
if RNum~length<3 then return RNum
if RNum~pos('.')>0 then do
    parse var RNum RNum '.' Fract
    Fract='.'||Fract
end
else Fract=''
RNum=RNum~reverse
FNum=''
do while RNum~length>3
	parse var RNum TriDig +3 RNum
	FNum=FNum||TriDig||','
end
FNum=FNum||RNum
return FNum~reverse||Fract
/*																	*/
/********************************************************************/

/* Display usage information ****************************************/
/*																	*/
::routine usage
say ''
say 'URL file fetcher v'||.rxiversion
say ''
say 'Usage:'
say ''
say 'urlget.cmd [options] <URL[;local name] | @filename> [options]'
say ''
say 'URL - URL of file to get'
say ''
say 'local name - local name to store file as'
say ''
say "@filename - text file with multiple URL's, one on a line"
say ''
say 'Options:'
say ''
say '/q - operate in quiet mode'
say '/r - attempt FTP resume if file exists and is smaller than server copy'
say '/f - attempt FTP resume, and abort transfer if not supported'
say "/n - don't overwrite local files"
say '/p - attempt transfer even if passive mode denied while behind firewall'
say '/w - act as if behind firewall (use passive mode)'
say '/d# - delay time for FTP commands (seconds); default=0.1'
say '/b# - transfer block size (512-65535); default=10240'
say '/t# - status timing interval in 1/100th seconds (1-100); default=25'
say '/m# - maximum retries on error (0-9999999999); default=1000'
say ''
say 'e.g. urlget http://www.somewhere.com/index.html'
return
/*																	*/
/********************************************************************/
